using Sandbox;
using System;

public class WalkControllerForce : BasePlayerController
{
    public float BodySize { get; set; } = 32.0f;
	
	public float BodyHeight { get; set; } = 72.0f;
	float Strength = 0.2f;
	public float strength { get { return Strength; } set { Strength = (value > 0) ? value : 0.2f; UpdateBox(); } }
	public float step;
	public float SwimmingSkill = 0.5f; //od 0 do 1
	public float MomentumSignificance { get; set; } = 1f;//0 to 1
	public float EyeHeight;
	public float MaxStandableAngle = 45.0f;
	public float gravity = 0.33f;

	public float GroundStrengthMultiplier{ get { return 1/FrictionGround - 1;} set { FrictionGround = 1 / (Math.Abs( value + 1 )); UpdateBox(); } }
	public float AirStrengthMultiplier{ get { return 1/FrictionAir - 1; } set { FrictionAir = 1 / (Math.Abs( value + 1 )); UpdateBox(); } }
	public float WaterStrengthMultiplier{ get { return 1/FrictionWater - 1; } set { FrictionWater = 1 / (Math.Abs( value + 1 )); UpdateBox(); } }
	float FrictionGround;
	float FrictionAir;
	float FrictionWater;


	TraceResult GroundRez, FrontRez;

	Trace tr1, tr2;

	Vector3 Vel, ResultVel;
	Vector3 Force = Vector3.Zero;
	float GroundAngle, MinimalSpeed;
	bool GroundExists, Swimming, Ducked = false;

	void DoDamage(float dmg){
		dmg = (dmg) / (1+Strength * 2);
		if(dmg > 10){
			Pawn.TakeDamage( DamageInfo.Generic(dmg));
		}
	}
	void UpdateBox(){
		var b = BodySize / 2.0f;
		tr1 = tr1.Size(new Vector3(-b, -b, 0), new Vector3(b,b,BodyHeight));
		EyeHeight = BodyHeight * 0.95f;
		step = BodyHeight / 5.0f;

		MinimalSpeed = Strength * (0.3f * (GroundStrengthMultiplier - 1) + 1);
	}

    public WalkControllerForce(){
		tr1 = Trace.Ray( 0, 0 )
		.WorldAndEntities()
		.HitLayer( CollisionLayer.All, false )
		.HitLayer( CollisionLayer.Solid, true )
		.HitLayer( CollisionLayer.GRATE, true )
		.HitLayer( CollisionLayer.PLAYER_CLIP, true );

		EyeHeight = BodyHeight * 0.95f;
		step = BodyHeight / 5.0f;
		UpdateBox();

		GroundStrengthMultiplier = 17f;
		AirStrengthMultiplier = 370;
		WaterStrengthMultiplier = 11f;
	}

	Vector3 ProjectOnPlane(Vector3 normal, Vector3 vector){
		return vector - normal * (Vector3.Dot( normal, vector ));
	}

	
	void SetFallDamage(Vector3 normal, float SolidNes){
		var damage = Vector3.Dot( normal, Force*MomentumSignificance ) * SolidNes;
		if(!Swimming || damage < 0)
			Force = ProjectOnPlane( normal, Force ) * SolidNes + Force * (1 - SolidNes);// + normal / 10.0f;
		DoDamage(-damage );
	}

	float GetSpeed(){
		if(Input.Down( InputButton.Run ))
			return Strength * 2f;
		if(Input.Down( InputButton.Walk ))
			return Strength * 0.5f;
		return Strength;
		
	}


	public override void Simulate()
	{
		tr2 = tr1;
		tr2.Ignore( Pawn );

		if ( Input.Pressed( InputButton.Duck ) && !Ducked && !Swimming)
		{
			BodyHeight /= 2.5f;
			Strength /= 2;
			UpdateBox();
			Ducked = true;
		}
		if ( Input.Released( InputButton.Duck ) && Ducked)
		{
			if(!(tr2.FromTo(Position + Vector3.Up * (BodyHeight), Position +  Vector3.Up * (step + BodyHeight * 2.5f)).Run().Hit))
			{

				BodyHeight *= 2.5f;
				Strength *= 2;
				UpdateBox();
				Ducked = false;
			}
		}


		EyeRot = Input.Rotation;
		EyePosLocal = Vector3.Up * EyeHeight;

		GroundRez = tr2.FromTo( Position, Position + Vector3.Down * 2 ).Run();
		GroundEntity = GroundRez.Entity;
		GroundExists = GroundEntity != null;
		GroundNormal = GroundRez.Normal;
		if(Swimming)
			Swimming = Pawn.WaterLevel.Fraction > 0.8f;
		else
		{
			Swimming = Pawn.WaterLevel.Fraction > 0.8f;
			if(Swimming)
				SetFallDamage( Vector3.Up, 1 - SwimmingSkill );
		}

		if ( Swimming )
		{
			var vertical = 0;
			if(Input.Down(InputButton.Jump)) vertical += 1;
			if(Input.Down( InputButton.Duck)) vertical -= 1;
			Vel = ((new Vector3( Input.Forward, Input.Left, 0 ) * EyeRot) + Vector3.Up * vertical).Normal * GetSpeed();
		}
		else
		{
			Vel = (new Vector3( Input.Forward, Input.Left, 0 ) * EyeRot).WithZ( 0 ).Normal * GetSpeed();

			// we dont want to slide like butter when beggining or stopping movement
	
			var len = Force.Length;
			if ( len > 0.1f && len < MinimalSpeed && Vector3.Dot( Force.Normal, Vel.Normal ) > 0.95f )
			{
				Force = Force.Normal * MinimalSpeed;
			}
			if ( (Force.DajDistance()) < MinimalSpeed && Vel == Vector3.Zero && GroundExists)
			{
				Force = 0;
			}
		}

	

		if ( !Swimming )
		{
			if ( GroundExists )
			{
				GroundAngle = GroundNormal.Angle( Vector3.Up );
				if ( GroundAngle != 0)
				{
					Vel += ProjectOnPlane( GroundNormal, Vector3.Down ) * Vel.Length * 1.3f;
					if ( GroundAngle > MaxStandableAngle )
					{
						Vel += Vector3.Down * gravity * 2;
						Vel = ProjectOnPlane( GroundNormal, Vel );
					}
				}
			}
			else
			{
				Vel = gravity * Vector3.Down * (Swimming ? 0.0f : 1);
			}
		}

		//Vel *= Time.Delta;
		Force += Vel;

		if(GroundExists)
			SetFallDamage( GroundNormal, 1 );

		if ( !Swimming )
		{
			if ( Input.Pressed( InputButton.Jump ) && GroundExists)
			{
				Force.z += Strength * 30f;
				Vel.z += Strength * 30f;
				GroundExists = false;
			}
			
			if ( GroundExists )
			{

				Force.x -= FrictionGround  * Force.x;
				Force.y -= FrictionGround  * Force.y;

				ResultVel = Force * MomentumSignificance + (1 - MomentumSignificance) * Vel;
				ResultVel += GroundEntity.BaseVelocity;
				WishVelocity = ResultVel;
				Velocity = WishVelocity;

				float done = 1;
				
				done = TryWalk( done );
				if(done > 0.05f)
					done = TryStepWalk( done );
				if ( done > 0.05f )
				{
					var predictedPos = Position + Force * MomentumSignificance;
					TryPartialWalk( done );
					DoDamage( (predictedPos - Position).Length );
					Force -= predictedPos - Position;
					Velocity = Force;
				}
				GoToGround();

			}
			else
			{
				Force -= FrictionAir * Force;

				ResultVel = Force * MomentumSignificance + (1 - MomentumSignificance) * Vel;
				WishVelocity = ResultVel;
				Velocity = WishVelocity;

				float done = 1;
				done = TryWalk( done );
				if ( done > 0.05f )
				{
					var predictedPos = Position + Force * MomentumSignificance;
					TryPartialWalk( done );
					DoDamage( (predictedPos - Position).Length );
					Force -= predictedPos - Position;
					Velocity = Force;
				}

			}
		}else //this is for swimming
		{
			Force -= FrictionWater * Force;

				ResultVel = Force * MomentumSignificance + (1 - MomentumSignificance) * Vel;
				WishVelocity = ResultVel;
				Velocity = WishVelocity;

				float done = 1;
				done = TryWalk( done );
				if ( done > 0.05f)
				{
					var predictedPos = Position + Force * MomentumSignificance;
					TryPartialWalk( done );
					DoDamage( (predictedPos - Position).Length );
					Force -= predictedPos - Position;
					Velocity = Force;
				}
		}




	}

	float TryWalk(float done){
		FrontRez = tr2.FromTo(Position, Position + ResultVel*done).Run();
		Position = FrontRez.EndPos;
		return done - FrontRez.Fraction;
	}

	float TryStepWalk(float done){
		if(tr2.FromTo(Position + Vector3.Up * (2), Position +  Vector3.Up * (step)).Run().Hit)
			return done;
		Position += Vector3.Up * step;
		if((Position - tr2.FromTo(Position, Position + ResultVel.Normal * BodySize).Run().EndPos).Length > BodySize / 3.0f)
			FrontRez = tr2.FromTo(Position, Position + ResultVel*done).Run();
		else{
			Position -= Vector3.Up * step;
			return done;
		}
		Position = FrontRez.EndPos;
		GoToGround();
		return done - FrontRez.Fraction*done;
	}

	float TryPartialWalk(float done){
		FrontRez = tr2.FromTo( Position, Position + ResultVel * done ).Run();
		var initial = ResultVel;
		Vector3[] normals = new Vector3[4];
		for ( int i = 0; i < 5; ++i){
			if(i == 4) return 1;
			normals[i] = FrontRez.Normal;
			if(i == 0)
				ResultVel = ProjectOnPlane( normals[0], ResultVel );
			else
			{
				ResultVel = initial;
				for ( int j = 0; j <= i; ++j )
					ResultVel = ProjectOnPlane( normals[j], ResultVel ) + normals[j]/20.0f;
				for ( int j = 0; j <= i;  ++j)
					if(ResultVel.Dot(normals[j]) > 0) break;
			}
			FrontRez = tr2.FromTo( Position, Position + ResultVel * done ).Run();
			Position = FrontRez.EndPos;
			done -= FrontRez.Fraction * done;
		}



		return done;
	}

	 void GoToGround(){
		Position = tr2.FromTo( Position, Position + Vector3.Down * step).Run().EndPos;

	}

}